!----------------------------------------------------------------------
!----------------------------------------------------------------------
! Copyright (C) 2004, 2005 Chaok Seok, Evangelos Coutsias and Ken Dill
!      UCSF, Univeristy of New Mexico, Seoul National University
! Witten by Chaok Seok and Evangelos Coutsias 2004.
                                                                                
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
                                                                                
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
                                                                                
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA!----------------------------------------------------------------------------
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
MODULE geometry
!-----------------------------------------------------------------------

  use ls_rmsd

  implicit none
  real(dp), parameter :: pi = 3.141592653589793238462643383279502884197d0
  real(dp), parameter :: deg2rad = pi/180.0d0, rad2deg = 180.0d0/pi
  real(dp), parameter :: ang_nac = 111.6*deg2rad, ang_acn = 117.5*deg2rad
  real(dp), parameter :: ang_cna = 120.0*deg2rad
  real(dp), parameter :: bnd_na = 1.45, bnd_ac = 1.52, bnd_cn = 1.33
  ! lbfgs-b parameters
  integer, parameter :: m_lbfgsb = 5 ! no of limited memory corrections stored
  integer, parameter :: mmax=17 !max no of limited memory corrections.
  integer :: max_iter_lbfgsb = 10000, iprint_lbfgsb = -1
  real(dp) :: factr = 1.0e+07, pgtol = 1.0d-3
  ! geometry variables
  integer, parameter :: max_res = 1000

!-----------------------------------------------------------------------
CONTAINS
!-----------------------------------------------------------------------
subroutine optimize_phi_psi(n_dof, num_res, coord_init, coord_opt, & 
     error_init, error_opt, U, center2, center1)
!-----------------------------------------------------------------------
! Use lbfgs-b for phi/psi optimization
!-----------------------------------------------------------------------
  integer, intent(in) :: n_dof, num_res
  real(dp), intent(in) :: coord_init(3,3*max_res)
  real(dp), intent(out) :: coord_opt(3,3*max_res)
  real(dp), intent(out) :: error_init, error_opt
  real(dp), intent(out) :: U(3,3), center1(3), center2(3)
  real(dp) :: coord(3,3*max_res), f
  integer :: N, status
  character(len=60) :: task, csave
  logical, dimension(4) :: lsave
  integer, dimension(44) :: isave
  real(dp), dimension(29) :: dsave
  integer :: iwa(3*n_dof), nbd(n_dof)
  real(dp) :: w_l(n_dof), w_u(n_dof), grad(n_dof), g(3,3*num_res)
  real(dp) :: wa(2*mmax*n_dof+4*n_dof+12*mmax*mmax+12*mmax), w(n_dof)
  logical :: calc_g

  N = 3*num_res

  nbd(1:n_dof) = 0 ! no bounds

  calc_g = .true.

  ! initial phi/psi
  coord(:,:) = coord_init(:,:)
  ! reconstruct backbone with canonical bond angles and omega
  call cartesian2internal(num_res, coord, w)
  call internal2cartesian(num_res, w, coord)
  call rmsd(N, coord, coord_init, 0, U, center2, center1, f, calc_g, g)
  call drmsd_dw(num_res, coord, g, grad)
  error_init = f

  task = 'START'   ! start the iteration by initializing task.
  !  L-BFGS-B loop 
  do
     ! This is the call to the L-BFGS-B code.
     call setulb(n_dof,m_lbfgsb,w,w_l,w_u,nbd,f,grad,factr,pgtol, &
          wa,iwa,task,iprint_lbfgsb,csave,lsave,isave,dsave)
     if (isave(30) > max_iter_lbfgsb) then
        if (iprint_lbfgsb >= 0) then
           write(*,*) 'Number of iteration exceeded maximum', max_iter_lbfgsb
           write(*,'(a6,2a12)') 'N_dof', 'Projg','F'
           write(*,'(i6,e12.3,f12.3)') isave(34), dsave(13), dsave(2)
        end if
        status = 1
        exit
     end if
     if (task(1:2) == 'FG') then
        ! the minimization routine has returned to request the
        ! function f and gradient g values at the current x.
        call internal2cartesian(num_res, w, coord)
        call rmsd(N, coord, coord_init, 1, U, center2, center1, f, calc_g, g)
        call drmsd_dw(num_res, coord, g, grad)
     endif
     if (task(1:5) /= 'NEW_X' .and. task(1:2) /= 'FG') exit
  end do

  if (task(1:4)=='ABNO' .or. task(1:5)=='ERROR') then
     if (iprint_lbfgsb >= 0) then
        write(*,*)'Abnormal termination or error in the lbfgs minimization.'
     end if
     status = 0
  else
     status = 1
  end if

  coord_opt(:,:) = coord(:,:)
  error_opt = f

  ! superpose to initial structure
  calc_g = .false.
  call rmsd(N, coord_opt, coord_init, 1, U, center2, center1, f, calc_g, g)

!-----------------------------------------------------------------------
end subroutine optimize_phi_psi
!-----------------------------------------------------------------------
subroutine test_drmsd_dw(num_res, coord, coord_init)
  integer, intent(in) :: num_res
  real(dp), intent(in) :: coord_init(3,3*max_res)
  real(dp), intent(out) :: coord(3,3*max_res)
  real(dp) :: U(3,3), center1(3), center2(3)
  real(dp) :: f, f0, dx, coord0(3,3*max_res), g0(3,3*num_res), g(3,3*num_res), w(2*num_res-2)
  logical :: calc_g
  integer :: res_no, idx, i, j

  calc_g = .true.

  ! initial phi/psi
  ! reconstruct backbone with canonical bond angles and omega
  coord(:,:) = coord_init(:,:)
  call cartesian2internal(num_res, coord, w)
  call internal2cartesian(num_res, w, coord)
  call rmsd(3*num_res, coord, coord_init, 1, U, center2, center1, f0, calc_g, g0)
!  call drmsd_dw(num_res, coord, g, grad)

  dx = 1.0d-3

  calc_g = .false.
  coord0(:,:) = coord(:,:)
  ! finite difference
  do res_no = 1, num_res
     idx = 3*(res_no - 1)
     do i = 1, 3
        do j = 1, 3
           coord(j,idx+i) = coord(j,idx+i) + dx
           call rmsd(3*num_res, coord, coord_init, 1, U, center2, center1, f, calc_g, g)
           write(*,"(2i5,3e15.5)") res_no, i, g0(j,idx+i), (f - f0)/dx, g0(j,idx+i)/((f - f0)/dx)
           coord(j,idx+i) = coord0(j,idx+i)
        end do
     end do
  end do

  stop
end subroutine test_drmsd_dw
!-----------------------------------------------------------------------
subroutine drmsd_dw(num_res, coord, g, grad)
  integer, intent(in) :: num_res
  real(dp), intent(in) :: coord(:,:), g(:,:)
  real(dp), intent(out) :: grad(:)
  integer :: ang_no, res_no, imin, imax, i, res_no1, i1min, i1, i_idx, res_idx
  real(dp) :: gi, r_pv(3), axis(3), dR(3), drdw(3)

  ang_no = 0
  do res_no = 1, num_res

     if (res_no == 1) then
        imin = 2
        imax = 2
     else if (res_no == num_res) then
        imin = 1
        imax = 1
     else
        imin = 1
        imax = 2
     end if

     res_idx = 3*(res_no-1)
     do i = imin, imax

        gi = 0.0d0

        if (i == 1) then
           ! phi
           r_pv(:) = coord(:,res_idx+2) ! CA
           axis(:) = coord(:,res_idx+2) - coord(:,res_idx+1)
        else
           ! psi
           r_pv(:) = coord(:,res_idx+3) ! C
           axis(:) = coord(:,res_idx+3) - coord(:,res_idx+2)
        end if

        ! axis of rotation
        axis(:) = axis(:)/sqrt(dot_product(axis,axis))

        do res_no1 = res_no, num_res

           if (res_no1 == res_no) then
              i1min = 3
           else 
              i1min = 1
           end if

           do i1 = i1min, 3
              i_idx = 3*(res_no1 - 1) + i1
              dR(:) = coord(:,i_idx) - r_pv(:)
              drdw(1)=axis(2)*dR(3)-axis(3)*dR(2)
              drdw(2)=axis(3)*dR(1)-axis(1)*dR(3)
              drdw(3)=axis(1)*dR(2)-axis(2)*dR(1)
              gi = gi + dot_product(g(:,i_idx),drdw(:))
           end do
        end do

        ang_no = ang_no + 1
        grad(ang_no) = gi
     end do
  end do

end subroutine drmsd_dw
!-----------------------------------------------------------------------
subroutine cartesian2internal(num_res, coord, w)
!-----------------------------------------------------------------------
! calc dih and bnd angles defined in "residue_ref" from bnd vectors
!-----------------------------------------------------------------------
  integer, intent(in) :: num_res
  real(dp), intent(in) :: coord(:,:)
  real(dp), intent(out) :: w(:)
  integer :: res_no, idx, i, ang_type = 1
  real(dp), dimension(3,3) :: r
  real(dp), dimension(3) :: r_n, r_a, r_c
  real(dp) :: phi, psi
  real(dp) :: bond(3,3,max_res)

  ! set bond vectors
  do res_no = 1, num_res
     idx = 3*(res_no - 1)
     r_n(:) = coord(:,idx+1)
     r_a(:) = coord(:,idx+2)
     r_c(:) = coord(:,idx+3)

     bond(:,1,res_no) = r_a(:) - r_n(:)
     bond(:,2,res_no) = r_c(:) - r_a(:)

     if (res_no < num_res) then
        bond(:,3,res_no) = coord(:,idx+4) - r_c(:)
     end if
  end do

  i = 0
  do res_no = 1, num_res
     ! phi
     if (res_no > 1) then
        r(:,1) = bond(:,3,res_no-1)
        r(:,2) = bond(:,1,res_no)
        r(:,3) = bond(:,2,res_no)
        call calc_angle(ang_type, r, phi)
        i = i + 1
        w(i) = phi
     end if
     ! psi
     if (res_no < num_res) then
        r(:,1) = bond(:,1,res_no)
        r(:,2) = bond(:,2,res_no)
        r(:,3) = bond(:,3,res_no)
        call calc_angle(ang_type, r, psi)
        i = i + 1
        w(i) = psi
     end if

  end do

end subroutine cartesian2internal
!-----------------------------------------------------------------------
subroutine internal2cartesian(num_res, w, coord)
!-----------------------------------------------------------------------
! calc cartesian coords from internal coords 
!-----------------------------------------------------------------------
  integer, intent(in) :: num_res
  real(dp), intent(in) :: w(:)
  real(dp), intent(out) :: coord(:,:)
  integer :: res_no, j, jmax, ang_type, ang_no, bnd_no, idx
  real(dp) :: q(4), quatn(4,6)
  real(dp) :: U(3), b(3), b_len
  real(dp) :: angle, angle_4,tan_w,tan_sqr,tan1, cosine,sine,p1,p2,p4
  real(dp) :: q1,q2,q3,q4

  ! arbitrary initial q and coord of first atm
  q(:) = (/ 1.0d0, 0.0d0, 0.0d0, 0.0d0 /)
  coord(:,1) = (/ 0.0d0, 0.0d0, 0.0d0 /) 

  ang_no = 0
  do res_no = 1, num_res

     if (res_no == num_res) then
        jmax = 4
     else
        jmax = 6
     end if

     do j = 1, jmax
        ! Place atom D, from angles ABCD and BCD
        if (mod(j,2) == 1) then
           ! dih ang
           ang_type = 1
           if (j == 1) then ! (-CA, -C, N, CA)
              ! omg
              angle = pi
           else if (j == 3) then ! (-C, N, CA, C)
              ! phi
              if (res_no == 1) then
                 angle = pi
              else
                 ang_no = ang_no + 1
                 angle = w(ang_no)
              end if
           else if (j == 5) then ! (N, CA, C, +N)
              ! psi
              if (res_no == num_res) then
                 angle = pi
              else
                 ang_no = ang_no + 1
                 angle = w(ang_no)
              end if
           end if
        else
           ! bnd ang
           ang_type = 2
           if (j == 2) then ! (-C, N, CA)
              angle = ang_cna
              b_len = bnd_na
           else if (j == 4) then ! (N, CA, C)
              angle = ang_nac
              b_len = bnd_ac
           else if (j == 6) then ! (CA, C, +N)
              angle = ang_acn
              b_len = bnd_cn
           end if
        end if

        ! calc quaternion and q product
        angle_4  = 0.25*angle
        tan_w = tan(angle_4)
        tan_sqr = tan_w * tan_w
        tan1 = 1.0d0 + tan_sqr
        cosine = (1.0d0 - tan_sqr)/tan1
        sine = 2.0d0*tan_w/tan1

        q1=q(1); q2=q(2); q3=q(3); q4=q(4)

        if (ang_type == 1) then
           p1 = cosine
           p2 = sine
           q(1) = q1*p1 - q2*p2; q(2) = q1*p2 + p1*q2
           q(3) = p1*q3 + q4*p2; q(4) = p1*q4 - q3*p2
        else if (ang_type == 2) then
           p1 = sine
           p4 = cosine
           q(1) = q1*p1 - q4*p4; q(2) = p1*q2 + q3*p4
           q(3) = p1*q3 - q2*p4; q(4) = p1*q4 + q1*p4
        end if

        if (ang_type == 2) then ! 1/2*U
           q1 = q(1); q2 = q(2); q3 = q(3); q4 = q(4)
           U(1) = q1*q1-0.5d0+q2*q2 
           U(2) = q2*q3+q1*q4 
           U(3) = q2*q4-q1*q3

           bnd_no = j/2
           b(:) = 2.0d0*b_len*U(:)
           idx = 3*(res_no - 1)+bnd_no
           coord(:,idx+1) = b + coord(:,idx)
        end if
     end do

  end do

end subroutine internal2cartesian
!-----------------------------------------------------------------------
subroutine calc_angle(ang_type, r, angle)
!-----------------------------------------------------------------------
! dih or imp: r1=Rab, r2=Rbc, r3=Rcd : angle between planes abc and bcd
! bnd ang: r1=Rab, r2=Rbc: angle between -r1 and r2
!-----------------------------------------------------------------------
  integer, intent(in) :: ang_type
  real(dp), dimension(:,:), intent(in) :: r
  real(dp), intent(out) :: angle
  real(dp), dimension(3) :: p, q, s
  real(dp) :: arg

  if (ang_type == 1 .or. ang_type ==3) then ! dihedral or improper
     call cross(r(:,1), r(:,2), p)
     call cross(r(:,2), r(:,3), q)
     call cross(r(:,3), r(:,1), s)
     arg = dot_product(p,q)/sqrt(dot_product(p,p)*dot_product(q,q))
     arg = sign(min(abs(arg),1.0d0),arg) ! to be sure abs(arg)<=1
     angle = sign(acos(arg), dot_product(s,r(:,2)))
  else if (ang_type == 2) then ! bond angle 
     p = -r(:,1); q = r(:,2)
     arg = dot_product(p,q)/sqrt(dot_product(p,p)*dot_product(q,q))
     arg = sign(min(abs(arg),1.0d0),arg) ! to be sure abs(arg)<=1
     angle = acos(arg)
  else
     write(*,*) 'Unknown angle type'
     stop
  end if
end subroutine calc_angle
!-----------------------------------------------------------------------
subroutine cross(p, q, s)
!-----------------------------------------------------------------------
  real(dp), dimension(:), intent(in) :: p, q
  real(dp), dimension(:), intent(out) :: s

  s(1)=p(2)*q(3)-p(3)*q(2)
  s(2)=p(3)*q(1)-p(1)*q(3)
  s(3)=p(1)*q(2)-p(2)*q(1)
end subroutine cross
!-----------------------------------------------------------------------
END MODULE geometry
!-----------------------------------------------------------------------
